package hexagon.graphics;

import java.nio.ByteBuffer;
import java.util.function.BiConsumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.GL31;
import org.lwjgl.opengl.GL33;

import hexagon.model.HexagonModel;
import hexagon.model.Loader;
import hexagon.world.Block;

public class BlockRenderer<T> {
	
	public static final BlockRenderer<Block> block, side;
	
	static {
		BiConsumer<Block, BlockRenderer<Block>> submitFunc = (b, r) -> {
			//r.buffer.putInt(b.x()).putInt(b.y()).putInt(b.z());
			r.buffer.putInt(b.blockType.texCoords);
			System.out.println("Deprecated (submitFunc)");
		};
		
		block = new BlockRenderer<>(3 * 4 + 2 * 4 + 3 * 4, 6, 3 * 4 + 1 * 4, 10000, submitFunc, (s, r) -> {
			int[] ints = { 1, 2, 0, 3, 5, 4 };
			for (int i = 0; i < 6; i++) {
				double v = ints[s == 0 ? i : 5 - i] * Math.PI / 3;
				float x = (float) Math.cos(v);
				float z = (float) Math.sin(v);
				r.buffer.putFloat(x).putFloat(1 - s).putFloat(z);
				r.buffer.putFloat((1 + x) / 2).putFloat((1 + z) / 2);
				r.buffer.putFloat(0).putFloat(1 - 2 * s).putFloat(0);
			}
		});
		
		side = new BlockRenderer<>(3 * 4 + 2 * 4 + 3 * 4, 4, 3 * 4 + 1 * 4, 10000, submitFunc, (s, r) -> {
			double nv = ((s - 1) % 6 - 0.5) * Math.PI / 3;
			float nx = (float) Math.cos(nv);
			float nz = (float) Math.sin(nv);
			for (int i = 0; i < 4; i++) {
				double v = (s - 1 - i % 2) % 6 * Math.PI / 3;
				float x = (float) Math.cos(v);
				float z = (float) Math.sin(v);
				r.buffer.putFloat(x).putFloat(1 - i / 2).putFloat(z);
				r.buffer.putFloat(1 - i % 2).putFloat(i / 2);
				r.buffer.putFloat(nx).putFloat(0).putFloat(nz);
			}
		});
	}
	
	
	private final int vertexSizeInBytes;
	private final int objectSizeInBytes;
	private final int verticesPerObject;
	public final int numberOfInstances;
	public final int instanceSize;
	private final int bufferSize;
	private final int vaoID, vboID;
	public final BiConsumer<T, BlockRenderer<T>> submitFunc;
	private final BiConsumer<Integer, BlockRenderer<T>> submitPerVertexFunc;
	
	private ByteBuffer buffer;
	private int instancesToDraw = 0;
	
	public BlockRenderer(int vertexSizeInBytes, int verticesPerObject, int instanceSize, int numberOfInstances, BiConsumer<T, BlockRenderer<T>> sumbitFunc, BiConsumer<Integer, BlockRenderer<T>> sumbitPerVertexFunc) {
		this.vertexSizeInBytes = vertexSizeInBytes;
		this.verticesPerObject = verticesPerObject;
		this.objectSizeInBytes = verticesPerObject * vertexSizeInBytes;
		this.numberOfInstances = numberOfInstances;
		this.instanceSize = instanceSize;
		this.bufferSize = this.objectSizeInBytes + this.numberOfInstances * this.instanceSize;
		this.submitFunc = sumbitFunc;
		this.submitPerVertexFunc = sumbitPerVertexFunc;
		vaoID = GL30.glGenVertexArrays();
		vboID = GL15.glGenBuffers();
		Loader.theLoader.vaos.add(vaoID);
		Loader.theLoader.vbos.add(vboID);
		
		GL30.glBindVertexArray(vaoID);
		buffer = BufferUtils.createByteBuffer(bufferSize);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, bufferSize, GL15.GL_STREAM_DRAW);
		GL20.glEnableVertexAttribArray(0);
		GL20.glEnableVertexAttribArray(1);
		GL20.glEnableVertexAttribArray(2);
		GL20.glEnableVertexAttribArray(3);
		GL20.glEnableVertexAttribArray(4);
		
		attribPointerFloat(0, 3, vertexSizeInBytes, 0); // position
		attribPointerFloat(1, 2, vertexSizeInBytes, 3 * 4); // texCoords
		attribPointerFloat(2, 3, vertexSizeInBytes, 5 * 4); // normals
		
		attribPointerInt(3, 3, instanceSize, objectSizeInBytes); // blockPos
		GL33.glVertexAttribDivisor(3, 1);
		attribPointerInt(4, 1, instanceSize, objectSizeInBytes + 3 * 4); // colorTint
		GL33.glVertexAttribDivisor(4, 1);
	}
	
	private void attribPointerInt(int index, int dims, int stride, int offset) {
		GL30.glVertexAttribIPointer(index, dims, GL11.GL_INT, stride, offset);
	}
	
	private void attribPointerFloat(int index, int dims, int stride, int offset) {
		GL20.glVertexAttribPointer(index, dims, GL11.GL_FLOAT, false, stride, offset);
	}
	
	/**	Usage: begin, submit ..., end, render */
	public void submit(T t) {
		if (buffer.remaining() < instanceSize) flush();
		
		submitFunc.accept(t, this);
	}
	
	public void submit(ByteBuffer data) {
		int pos = 0, limit = data.position();
		
		while (pos < limit) {
			int write = Math.min(limit - pos, buffer.remaining());
			buffer.put(data.array(), pos, write);
			flush();
			pos += write;
		}
	}
	
	public void submitPerVertexData(int side) {
		buffer.position(0);
		submitPerVertexFunc.accept(side, this);
		buffer.position(objectSizeInBytes);
	}
	
	public void flush() {
		end();
		render();
		begin();
	}
	
	public void begin() {
		buffer.clear();
		buffer.position(objectSizeInBytes);
	}
	
	public void end() {
		instancesToDraw = (buffer.position() - objectSizeInBytes) / instanceSize;
		buffer.flip();
//		System.out.println("lim: " + buffer.limit());
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, buffer.limit(), GL15.GL_STREAM_DRAW);
		GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, buffer);
	}
	
	public void render() {
		GL30.glBindVertexArray(vaoID);
		GL13.glActiveTexture(GL13.GL_TEXTURE0);
		GL11.glBindTexture(GL30.GL_TEXTURE_2D_ARRAY, HexagonModel.theModel.getTextureID());
		
		GL31.glDrawArraysInstanced(GL11.GL_TRIANGLE_STRIP, 0, verticesPerObject, instancesToDraw);
	}
	
}
