package hexagon.world;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;

import hexagon.world.block.BlockType;

public class Chunk {
	/** Number of chunks in a ring */
	public static final int clusterCircumference = 16;//64k MAX, 16k works, 4k acceptable, 1k good, 256 excellent
	public static final int ringSize = 16 * clusterCircumference;
	public static final int totalSize = 16 * ringSize;
	
	protected Cluster cluster;
	protected short[] blocks;
	protected byte[] sidesToRender;
	protected boolean[] willUpdate;
	public final short coords;
	
	public Chunk(int x, int y, int z, Cluster cluster) {
		this(encodeCoords(x, y, z), cluster, false);
	}
	
	public Chunk(int x, int y, int z, Cluster cluster, boolean genNew) {
		this(encodeCoords(x, y, z), cluster, genNew);
	}
	
	public Chunk(short coords, Cluster cluster) {
		this(coords, cluster, false);
	}
	
	public Chunk(short coords, Cluster cluster, boolean genNew) {
		this.coords = coords;
		this.cluster = cluster;
		
		int size = 16 * 16 * 16;
		blocks = new short[size];
		sidesToRender = new byte[size];
		willUpdate = new boolean[size];
		
		if (genNew || true) {//!loadFromFile()) {
			cluster.world.landscapeGen.genChunk(this);
		}
	}
	
	public int localX() {
		return coords >> 8 & 15;
	}
	
	public int localY() {
		return coords >> 4 & 15;
	}
	
	public int localZ() {
		return coords & 15;
	}
	
	public int x() {
		return cluster.x() << 4 | coords >> 8 & 15;
	}
	
	public int y() {
		return cluster.y() << 4 | coords >> 4 & 15;
	}
	
	public int z() {
		return cluster.z() << 4 | coords & 15;
	}
	
	public int bx(short coords) {
		return x() << 4 | coords >> 8 & 15;
	}
	
	public int by(short coords) {
		return y() << 4 | coords >> 4 & 15;
	}
	
	public int bz(short coords) {
		return z() << 4 | coords & 15;
	}
	
	public int bLocalX(short coords) {
		return coords >> 8 & 15;
	}
	
	public int bLocalY(short coords) {
		return coords >> 4 & 15;
	}
	
	public int bLocalZ(short coords) {
		return coords & 15;
	}
	
	public long worldCoords() {
		return cluster.coords << 12 | coords;
	}
	
	public BlockType getBlock(short block) {
		return blocks != null ? BlockType.blockTypes.get(blocks[block]) : null;
	}
	
	public BlockType getBlock(int x, int y, int z) {
		return blocks != null ? BlockType.blockTypes.get(blocks[encodeCoords(x, y, z)]) : null;
	}
	
	public short setBlock(short bCoords, BlockType block) {
		short b = blocks[bCoords];
		blocks[bCoords] = block.id;
		cluster.world.addBlockToUpdateList(this, bCoords);
		return b;
	}
	
	public short setBlock(int x, int y, int z, BlockType block) {
		return setBlock(encodeCoords(x, y, z), block);
	}
	
	public short removeBlock(int x, int y, int z) {
		short coords = encodeCoords(x, y, z);
		short b = blocks[coords];
		blocks[coords] = 0;
		
		onRemoveBlock(coords, false);
		
		return b;
	}
	
	public void onRemove(boolean noUpdate) {
		saveChunk();
		
		if (blocks != null) {
			for (short coords = 0; coords < blocks.length; coords++) {
				if (blocks[coords] == 0) continue;
				
				onRemoveBlock(coords, false);//TODO: noUpdate
			}
			blocks = null;
		}
		cluster = null;
	}
	
	private void onRemoveBlock(short bCoords, boolean noUpdate) {
		if (!noUpdate) {
			long wCoords = World.encodeBlockCoords(cluster.coords, this.coords, bCoords);
			if (sidesToRender[bCoords] != 0) {
				for (int i = 0; i < 8; i++) {
					if ((sidesToRender[bCoords] & 1 << i) != 0) cluster.world.removeFromRenderList(i, wCoords);
				}
			}
			sidesToRender[bCoords] = 0;
			
			BlockType b;
			for (int i = 0; i < 8; i++) {
				long coords = World.encodeBlockCoords(bx(bCoords) + Block.neighbourOffsets[i][0], by(bCoords) + Block.neighbourOffsets[i][1], bz(bCoords) + Block.neighbourOffsets[i][2]);
				b = cluster.world.getBlock(coords);
				if (b != null) cluster.world.addBlockToUpdateList(coords);
			}
		}
	}
	
	public void onUpdateBlock(short inCoords) {
		byte sides = 0;
		Chunk c;
		BlockType b;
		short bCoords = 0;
		for (int i = 0; i < 8; i++) {
			long coords = World.encodeBlockCoords(bx(inCoords) + Block.neighbourOffsets[i][0], by(inCoords) + Block.neighbourOffsets[i][1], bz(inCoords) + Block.neighbourOffsets[i][2]);
			c = cluster.world.getChunk(coords >>> 12);
			b = c == null ? null : c.getBlock(bCoords = (short) (coords & 0xfff));
			if (b == null) sides |= 1 << i;
			else if ((c.sidesToRender[bCoords] & 1 << (i < 2 ? 1 - i : (i - 2 + 3) % 6 + 2)) != 0) cluster.world.addBlockToUpdateList(c, bCoords);
		}
		sidesToRender[inCoords] = sides;
	}
	
	public static short encodeCoords(int x, int y, int z) { // bits in short: 4x 4y 4z
		return (short) ((x & 0xf) << 8 | (y & 0xf) << 4 | z & 0xf);
	}
	
	protected void saveChunk() {
		return;
/*		if (cluster.world.saveFolder != null && blocks != null) {
//			cluster.world.bytes = 0;
//			cluster.world.time = 0;
			cluster.world.saveBuffer.clear();
			try {
				File file = new File(cluster.world.saveFolder, "chunks");
				file.mkdirs();
				file = new File(file, x() + "." + y() + "." + z() + ".dat");
				if (file.exists()) file.delete();
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
				cluster.world.saveBuffer.put("HX".getBytes());// H X maj min
				cluster.world.saveBuffer.put((byte) 0);
				cluster.world.saveBuffer.put((byte) 1);
				
				for (Block b : blocks)
					if (b != null) b.saveBlock(cluster.world.saveBuffer, fc);
				
				cluster.world.saveBufferFull.accept(fc);
				
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			
//			System.out.println(cluster.world.bytes + " bytes written in " + cluster.world.time + " ns. That's " + (int) (cluster.world.bytes / 8e3 / (cluster.world.time / 1e9) * 10) / 10f + " kB/s");
		}*/
	}
	
	/** Tries to load from file. @return Whether or not the file existed and the loading could be performed. */
	private boolean loadFromFile() {//TODO: improve performance
		if (cluster.world.saveFolder != null) {
//			cluster.world.bytes = 0;
			cluster.world.readBuffer.clear();
//			long start = System.nanoTime();
			try {
				File file = new File(cluster.world.saveFolder, "chunks/" + x() + "." + y() + "." + z() + ".dat");
				if (!file.exists()) return false;
				RandomAccessFile rFile = new RandomAccessFile(file, "r");
				FileChannel fc = rFile.getChannel();
				
				cluster.world.readBuffer.position(cluster.world.readBuffer.limit() - 4);
				fc.read(cluster.world.readBuffer);
				cluster.world.readBuffer.position(cluster.world.readBuffer.limit() - 4);
				
				if (cluster.world.readBuffer.get() == 'H' && cluster.world.readBuffer.get() == 'X') {
					short version = cluster.world.readBuffer.getShort();
					cluster.world.readBuffer.clear();
					
					switch (version >> 8 & 15) {
						case 0:
							switch (version & 15) {
								case 1:
									while (fc.read(cluster.world.readBuffer) > 0) {
										cluster.world.readBuffer.flip();
//										cluster.world.bytes += cluster.world.readBuffer.remaining();
										while (cluster.world.readBuffer.remaining() >= 4) {
											short coords = cluster.world.readBuffer.getShort();
											if (coords < 0) System.out.println("Nooo!");
											short blockType = cluster.world.readBuffer.getShort();
											setBlock(coords, BlockType.blockTypes.get(blockType));
										}
										cluster.world.readBuffer.clear();
									}
									break;
								default:
									System.err.println("Chunkfile has an unsupported minor version: " + (version & 15) + ".\n\tFilepath: " + file.getAbsolutePath());
							}
							break;
						default:
							System.err.println("Chunkfile has an unsupported major version: " + (version >>> 8) + ".\n\tFilepath: " + file.getAbsolutePath());
					}
				} else {
					System.err.println("Chunkfile is not a valid chunkfile.\n\tFilepath: " + file.getAbsolutePath());
				}
				fc.close();
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
//			start = System.nanoTime() - start;
//			System.out.println(cluster.world.bytes + " bytes read in " + start + " ns. That's " + (int) (cluster.world.bytes / 8e3 / (start / 1e9) * 10) / 10f + " kB/s");
			
			return true;
		}
		return false;
	}
	
}
