package hexagon.graphics;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL32;
import org.lwjgl.opengl.GL40;

import hexagon.Main;

public class ShaderBuilder {
	
	private HashMap<Integer, Integer> shaders;
	private int programID;
	private String name, prefix = "res/shaders/";
	
	private ShaderBuilder(String name) {
		this.name = name;
		shaders = new HashMap<>(3);
		programID = GL20.glCreateProgram();
	}
	
	public static ShaderBuilder start(String name) {
		return new ShaderBuilder(name);
	}
	
	public ShaderBuilder setPrefix(String prefix) {
		this.prefix = prefix;
		return this;
	}
	
	public ShaderBuilder load(String type) {
		return load(type, prefix + name + '.' + type);
	}
	
	public ShaderBuilder load(String type, String path) {
		int t;
		switch (type) {
			case "vs":
			case "vert":
				t = GL20.GL_VERTEX_SHADER;
				break;
			case "fs":
			case "frag":
				t = GL20.GL_FRAGMENT_SHADER;
				break;
			case "gs":
			case "geom":
				t = GL32.GL_GEOMETRY_SHADER;
				break;
			case "tc":
				t = GL40.GL_TESS_CONTROL_SHADER;
				break;
			case "te":
				t = GL40.GL_TESS_EVALUATION_SHADER;
				break;
			default:
				System.err.println("Shadertype '" + type + "' not supported.");
				return this;
		}
		return load(t, path);
	}
	
	public ShaderBuilder load(int type, String path) {
		int shaderID = GL20.glCreateShader(type);
		StringBuilder source = new StringBuilder();
		
		try {
			BufferedReader reader = new BufferedReader(new FileReader(path));
			String line;
			while ((line = reader.readLine()) != null)
				source.append(line).append('\n');
			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
			Main.destroy();
			System.exit(1);
		}
		
		GL20.glShaderSource(shaderID, source);
		GL20.glCompileShader(shaderID);
		if (GL20.glGetShaderi(shaderID, GL20.GL_COMPILE_STATUS) == GL11.GL_FALSE) System.err.println((type == GL20.GL_VERTEX_SHADER ? "Vertexshader" : type == GL20.GL_FRAGMENT_SHADER ? "Fragmentshader" : "Shader") + " failed to compile.\nError log:\n" + GL20.glGetShaderInfoLog(shaderID, GL20.glGetShaderi(shaderID, GL20.GL_INFO_LOG_LENGTH)));
		shaders.put(type, shaderID);
		return this;
	}
	
	public ShaderBuilder bindAttribs(String... attribs) {
		int index = 0;
		for (String s : attribs)
			GL20.glBindAttribLocation(programID, index++, s);
		return this;
	}
	
	public ShaderBuilder attatchAll() {
		for (int i : shaders.values())
			GL20.glAttachShader(programID, i);
		return this;
	}
	
	public Shader linkAndFinish() {
		GL20.glLinkProgram(programID);
		
		if (GL20.glGetProgrami(programID, GL20.GL_LINK_STATUS) == GL11.GL_FALSE) {
			System.err.println("Link error: " + GL20.glGetProgramInfoLog(programID, GL20.glGetShaderi(programID, GL20.GL_INFO_LOG_LENGTH)));
		}
		
		GL20.glValidateProgram(programID);
		
		if (GL20.glGetProgrami(programID, GL20.GL_VALIDATE_STATUS) == GL11.GL_FALSE) {
			System.err.println("Validation error: " + GL20.glGetProgramInfoLog(programID, GL20.glGetShaderi(programID, GL20.GL_INFO_LOG_LENGTH)));
		}
		
		for (int i : shaders.values()) {
			GL20.glDetachShader(programID, i);
			GL20.glDeleteShader(i);
		}
		
		return new Shader(programID);
	}
	
}
