package hexagon.world;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.util.Arrays;

import hexagon.world.block.BlockType;

public class Chunk {
	/** Number of blocks in a chunk in one dimension (eg. 4 means 4x4x4 blocks).<br><b>Do not change!</b> */
	public static final int chunkSize = 16;
	/** Number of chunks in a ring */
	public static final int ringSize = 256;//256k works, 64k acceptable, 16k good, 4k excellent
	public static final int totalSize = chunkSize * ringSize;
	
	protected World world;
	protected Block[] blocks;
	public final int x, y, z;
	
	public Chunk(int x, int y, int z, World world) {
		this(x, y, z, world, false);
	}
	
	public Chunk(int x, int y, int z, World world, boolean genNew) {
		this.x = x;
		this.y = y;
		this.z = z;
		this.world = world;
		
		if (genNew || !loadFromFile()) {
			if (y == -1) {// the rest is not implemented yet!
				for (int i = 0; i < Chunk.chunkSize; i++) {
					for (int j = 0; j < Chunk.chunkSize; j++) {
						for (int k = 10; k < 14; k++) {
							setBlock(new Block(i, k, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
						}
						if (Math.random() < 0.9) setBlock(new Block(i, 14, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
						if (Math.random() < 0.8) setBlock(new Block(i, 15, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
					}
				}
			}
		}
	}
	
	public Block getBlock(short block) {
		return blocks != null ? blocks[block] : null;
	}
	
	public Block getBlock(int x, int y, int z) {
		return blocks != null ? blocks[encodeCoords(x, y, z)] : null;
	}
	
	public Block setBlock(Block block) {
		long coords = World.encodeBlockCoords(x, y, z, block.coords);
		world.addBlockToUpdateList(coords, block);
		if (blocks == null) blocks = new Block[16 * 16 * 16];
		return blocks[block.coords] = block;
	}
	
	public Block removeBlock(int x, int y, int z) {
		short coords = encodeCoords(x, y, z);
		long wCoords = World.encodeBlockCoords(this.x, this.y, this.z, coords);
		Block b = blocks[coords];
		blocks[coords] = null;
		
		if (b.sidesToRender != 0) {
			for (int i = 0; i < 8; i++) {
				if ((b.sidesToRender & 1 << i) != 0)
					world.removeFromRenderList(i, wCoords);
			}
		}
		
		b.onRemove(false);
		
		return b;
	}
	
	public void onRemove(boolean noUpdate) {
		saveChunk();
		
		if (blocks != null) {
			long thisChunk = World.encodeCoords(x, y, z);
			Block b;
			for (short coords = 0; coords < blocks.length; coords++) {
				if ((b = blocks[coords]) == null) continue;
				
				if (b.sidesToRender != 0) {
					for (int i = 0; i < 8; i++) {
						if ((b.sidesToRender & 1 << i) != 0)
							world.removeFromRenderList(i, World.encodeBlockCoords(thisChunk, coords));
					}
				}
				
				b.onRemove(false);
			}
			Arrays.fill(blocks, null);
		}
		world = null;
	}
	
	public static short encodeCoords(int x, int y, int z) { // bits in short: 4x 4y 4z
		return (short) ((x & 0xf) << 8 | (y & 0xf) << 4 | z & 0xf);
	}
	
	protected void saveChunk() {
		if (world.saveFolder != null && blocks != null) {
//			world.bytes = 0;
//			world.time = 0;
			world.saveBuffer.clear();
			try {
				File file = new File(world.saveFolder, "chunks");
				file.mkdirs();
				file = new File(file, x + "." + y + "." + z + ".dat");
				if (file.exists()) file.delete();
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
				world.saveBuffer.put("HX".getBytes());// H X maj min
				world.saveBuffer.put((byte) 0);
				world.saveBuffer.put((byte) 1);
				
				for (Block b : blocks)
					if (b != null) b.saveBlock(world.saveBuffer, fc);
				
				world.saveBufferFull.accept(fc);
				
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			
//			System.out.println(world.bytes + " bytes written in " + world.time + " ns. That's " + (int) (world.bytes / 8e3 / (world.time / 1e9) * 10) / 10f + " kB/s");
		}
	}
	
	/** Tries to load from file. @return Whether or not the file existed and the loading could be performed. */
	private boolean loadFromFile() {//TODO: improve performance
		if (world.saveFolder != null) {
//			world.bytes = 0;
			world.readBuffer.clear();
//			long start = System.nanoTime();
			try {
				File file = new File(world.saveFolder, "chunks/" + x + "." + y + "." + z + ".dat");
				if (!file.exists()) return false;
				RandomAccessFile rFile = new RandomAccessFile(file, "r");
				FileChannel fc = rFile.getChannel();
				
				world.readBuffer.position(world.readBuffer.limit() - 4);
				fc.read(world.readBuffer);
				world.readBuffer.position(world.readBuffer.limit() - 4);
				
				if (world.readBuffer.get() == 'H' && world.readBuffer.get() == 'X') {
					short version = world.readBuffer.getShort();
					world.readBuffer.clear();
					
					switch (version >> 8 & 15) {
						case 0:
							switch (version & 15) {
								case 1:
									while (fc.read(world.readBuffer) > 0) {
										world.readBuffer.flip();
//										world.bytes += world.readBuffer.remaining();
										while (world.readBuffer.remaining() >= 4) {
											short coords = world.readBuffer.getShort();
											if (coords < 0) System.out.println("Nooo!");
											short blockType = world.readBuffer.getShort();
											setBlock(new Block(coords, BlockType.blockTypes.get(blockType), this));
										}
										world.readBuffer.clear();
									}
									break;
								default:
									System.err.println("Chunkfile has an unsupported minor version: " + (version & 15) + ".\n\tFilepath: " + file.getAbsolutePath());
							}
							break;
						default:
							System.err.println("Chunkfile has an unsupported major version: " + (version >>> 8) + ".\n\tFilepath: " + file.getAbsolutePath());
					}
				} else {
					System.err.println("Chunkfile is not a valid chunkfile.\n\tFilepath: " + file.getAbsolutePath());
				}
				fc.close();
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
//			start = System.nanoTime() - start;
//			System.out.println(world.bytes + " bytes read in " + start + " ns. That's " + (int) (world.bytes / 8e3 / (start / 1e9) * 10) / 10f + " kB/s");
			
			return true;
		}
		return false;
	}
	
}
