package hexagon.world;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.util.Arrays;

import hexagon.world.block.BlockType;

public class Chunk {
	/** Number of chunks in a ring */
	public static final int ringSize = 256;//256k works, 64k acceptable, 16k good, 4k excellent
	public static final int totalSize = 16 * ringSize;
	
	protected Cluster cluster;
	protected Block[] blocks;
	public final short coords;
	
	public Chunk(int x, int y, int z, Cluster cluster) {
		this(x, y, z, cluster, false);
	}
	
	public Chunk(int x, int y, int z, Cluster cluster, boolean genNew) {
		this.coords = encodeCoords(x, y, z);
		this.cluster = cluster;
		
		if (genNew || !loadFromFile()) {
			if (y() == -1) {// the rest is not implemented yet!
				for (int i = 0; i < 16; i++) {
					for (int j = 0; j < 16; j++) {
						for (int k = 10; k < 14; k++) {
							setBlock(new Block(i, k, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
						}
						if (Math.random() < 0.9) setBlock(new Block(i, 14, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
						if (Math.random() < 0.8) setBlock(new Block(i, 15, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
					}
				}
			}
		}
	}
	
	public int localX() {
		return coords >> 8 & 15;
	}
	
	public int localY() {
		return coords >> 4 & 15;
	}
	
	public int localZ() {
		return coords & 15;
	}
	
	public int x() {
		return cluster.x() << 4 | coords >> 8 & 15;
	}
	
	public int y() {
		return cluster.y() << 4 | coords >> 4 & 15;
	}
	
	public int z() {
		return cluster.z() << 4 | coords & 15;
	}
	
	public long worldCoords() {
		return cluster.coords << 12 | coords;
	}
	
	public Block getBlock(short block) {
		return blocks != null ? blocks[block] : null;
	}
	
	public Block getBlock(int x, int y, int z) {
		return blocks != null ? blocks[encodeCoords(x, y, z)] : null;
	}
	
	public Block setBlock(Block block) {
		long coords = World.encodeBlockCoords(cluster.coords, this.coords, block.coords);
		cluster.world.addBlockToUpdateList(coords, block);
		if (blocks == null) blocks = new Block[16 * 16 * 16];
		Block b = blocks[block.coords];
		blocks[block.coords] = block;
		if (b != null) b.onRemove(false);
		return block;
	}
	
	public Block removeBlock(int x, int y, int z) {
		short coords = encodeCoords(x, y, z);
		long wCoords = World.encodeBlockCoords(cluster.coords, this.coords, coords);
		Block b = blocks[coords];
		blocks[coords] = null;
		
		if (b.sidesToRender != 0) {
			for (int i = 0; i < 8; i++) {
				if ((b.sidesToRender & 1 << i) != 0)
					cluster.world.removeFromRenderList(i, wCoords);
			}
		}
		
		b.onRemove(false);
		
		return b;
	}
	
	public void onRemove(boolean noUpdate) {
		saveChunk();
		
		if (blocks != null) {
			Block b;
			for (short coords = 0; coords < blocks.length; coords++) {
				if ((b = blocks[coords]) == null) continue;
				
				if (b.sidesToRender != 0) {
					for (int i = 0; i < 8; i++) {
						if ((b.sidesToRender & 1 << i) != 0)
							cluster.world.removeFromRenderList(i, World.encodeBlockCoords(cluster.coords, this.coords, coords));
					}
				}
				
				b.onRemove(false);//TODO: noUpdate
			}
			Arrays.fill(blocks, null);
		}
		cluster = null;
	}
	
	public static short encodeCoords(int x, int y, int z) { // bits in short: 4x 4y 4z
		return (short) ((x & 0xf) << 8 | (y & 0xf) << 4 | z & 0xf);
	}
	
	protected void saveChunk() {
		if (cluster.world.saveFolder != null && blocks != null) {
//			cluster.world.bytes = 0;
//			cluster.world.time = 0;
			cluster.world.saveBuffer.clear();
			try {
				File file = new File(cluster.world.saveFolder, "chunks");
				file.mkdirs();
				file = new File(file, x() + "." + y() + "." + z() + ".dat");
				if (file.exists()) file.delete();
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
				cluster.world.saveBuffer.put("HX".getBytes());// H X maj min
				cluster.world.saveBuffer.put((byte) 0);
				cluster.world.saveBuffer.put((byte) 1);
				
				for (Block b : blocks)
					if (b != null) b.saveBlock(cluster.world.saveBuffer, fc);
				
				cluster.world.saveBufferFull.accept(fc);
				
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			
//			System.out.println(cluster.world.bytes + " bytes written in " + cluster.world.time + " ns. That's " + (int) (cluster.world.bytes / 8e3 / (cluster.world.time / 1e9) * 10) / 10f + " kB/s");
		}
	}
	
	/** Tries to load from file. @return Whether or not the file existed and the loading could be performed. */
	private boolean loadFromFile() {//TODO: improve performance
		if (cluster.world.saveFolder != null) {
//			cluster.world.bytes = 0;
			cluster.world.readBuffer.clear();
//			long start = System.nanoTime();
			try {
				File file = new File(cluster.world.saveFolder, "chunks/" + x() + "." + y() + "." + z() + ".dat");
				if (!file.exists()) return false;
				RandomAccessFile rFile = new RandomAccessFile(file, "r");
				FileChannel fc = rFile.getChannel();
				
				cluster.world.readBuffer.position(cluster.world.readBuffer.limit() - 4);
				fc.read(cluster.world.readBuffer);
				cluster.world.readBuffer.position(cluster.world.readBuffer.limit() - 4);
				
				if (cluster.world.readBuffer.get() == 'H' && cluster.world.readBuffer.get() == 'X') {
					short version = cluster.world.readBuffer.getShort();
					cluster.world.readBuffer.clear();
					
					switch (version >> 8 & 15) {
						case 0:
							switch (version & 15) {
								case 1:
									while (fc.read(cluster.world.readBuffer) > 0) {
										cluster.world.readBuffer.flip();
//										cluster.world.bytes += cluster.world.readBuffer.remaining();
										while (cluster.world.readBuffer.remaining() >= 4) {
											short coords = cluster.world.readBuffer.getShort();
											if (coords < 0) System.out.println("Nooo!");
											short blockType = cluster.world.readBuffer.getShort();
											setBlock(new Block(coords, BlockType.blockTypes.get(blockType), this));
										}
										cluster.world.readBuffer.clear();
									}
									break;
								default:
									System.err.println("Chunkfile has an unsupported minor version: " + (version & 15) + ".\n\tFilepath: " + file.getAbsolutePath());
							}
							break;
						default:
							System.err.println("Chunkfile has an unsupported major version: " + (version >>> 8) + ".\n\tFilepath: " + file.getAbsolutePath());
					}
				} else {
					System.err.println("Chunkfile is not a valid chunkfile.\n\tFilepath: " + file.getAbsolutePath());
				}
				fc.close();
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
//			start = System.nanoTime() - start;
//			System.out.println(cluster.world.bytes + " bytes read in " + start + " ns. That's " + (int) (cluster.world.bytes / 8e3 / (start / 1e9) * 10) / 10f + " kB/s");
			
			return true;
		}
		return false;
	}
	
}
